/**
 * Copyright (c) 2013 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**@file
 *
 * @defgroup ble_sdk_app_connectivity_main main.c
 * @{
 * @ingroup ble_sdk_app_connectivity
 *
 * @brief BLE Connectivity application.
 */

#include <stdbool.h>
#include "nrf_sdm.h"
#include "nrf_soc.h"
#include "app_error.h"
#include "app_scheduler.h"
#include "nrf_sdh.h"
#include "ser_hal_transport.h"
#include "ser_conn_handlers.h"
#include "boards.h"
#include "nrf_drv_clock.h"

#define NRF_LOG_LEVEL 4
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "app_timer.h"
#include "ser_phy_debug_comm.h"

#include "app_usbd_cdc_acm.h"
#include "my_uart.h"
#include "fds.h"
#include "fds_example.h"
#include "nrf_delay.h"

#if defined(APP_USBD_ENABLED) && APP_USBD_ENABLED
#include "app_usbd_serial_num.h"
#ifdef BOARD_PCA10059
#include "nrf_dfu_trigger_usb.h"
#endif
#include "app_usbd.h"

#include "MegaSig_BLE_Function.h"
#include "MegaSig_BLE_Gattc.h"
#include "MegaSig_Global.h"

static volatile bool m_usb_started;

static void usbd_user_evt_handler(app_usbd_event_type_t event)
{
  switch (event)
  {
  case APP_USBD_EVT_DRV_SUSPEND:
    break;
  case APP_USBD_EVT_DRV_RESUME:
    break;
  case APP_USBD_EVT_STARTED:
    m_usb_started = true;
    break;
  case APP_USBD_EVT_STOPPED:
    app_usbd_disable();
    break;
  case APP_USBD_EVT_POWER_DETECTED:
    NRF_LOG_INFO("USB power detected");

    if (!nrf_drv_usbd_is_enabled())
    {
      app_usbd_enable();
    }
    break;
  case APP_USBD_EVT_POWER_REMOVED:
    NRF_LOG_INFO("USB power removed");
    app_usbd_stop();
    break;
  case APP_USBD_EVT_POWER_READY:
    NRF_LOG_INFO("USB ready");
    app_usbd_start();
    break;
  default:
    break;
  }
}
/*
static uint32_t m_connect()
{
  uint32_t err_code;
  // uint8_t address[6] = {0x5d, 0xd3, 0x15, 0xc4, 0x79, 0xe0};
  ble_gap_addr_t addr = {1, 2, {0x5d, 0xd3, 0x15, 0xc4, 0x79, 0xe0}};
  // e079c415d35d

  err_code = sd_ble_gap_connect(&addr, &m_scan_param, &m_connection_param, 1);
  if (err_code != NRF_SUCCESS)
  {
    NRF_LOG_DEBUG("sd_ble_gap_connect: 0x%x", err_code);
  }

  return err_code;
}

ble_data_t adv_report_data;
uint8_t mp_data[100] = {0};
static uint32_t m_scan_start()
{
  adv_report_data.p_data = mp_data;
  adv_report_data.len = sizeof(mp_data);

  uint32_t err_code = sd_ble_gap_scan_start(&m_scan_param, &adv_report_data);
  NRF_LOG_INFO("sd_ble_gap_scan_start: 0x%x\r", err_code);
  return err_code;
} */

static void usbd_init(void)
{
  app_usbd_serial_num_generate();
  static const app_usbd_config_t usbd_config = {
      .ev_state_proc = usbd_user_evt_handler};
  APP_ERROR_CHECK(app_usbd_init(&usbd_config));
}

static void usbd_enable(void)
{
#ifdef BOARD_PCA10059
  APP_ERROR_CHECK(nrf_dfu_trigger_usb_init());
#endif
  APP_ERROR_CHECK(app_usbd_power_events_enable());

  /* Process USB events until USB is started. This is related to the fact that
   * current version of softdevice does not handle USB POWER events. */
  while (m_usb_started == false)
  {
    while (app_usbd_event_queue_process())
    {
      /* Nothing to do */
    }
  }
}
#endif // APP_USBD_ENABLED

static void on_idle(void)
{

  if (!NRF_LOG_PROCESS())
  {
    // Wait for an event.
    if (nrf_sdh_is_enabled())
    {
      ret_code_t ret_code = sd_app_evt_wait();
      ASSERT((ret_code == NRF_SUCCESS) || (ret_code == NRF_ERROR_SOFTDEVICE_NOT_ENABLED));
      UNUSED_VARIABLE(ret_code);
    }
    else
    {
      // Wait for an event.
      __WFE();
      // Clear the internal event register.
      __SEV();
      __WFE();
    }
  }

#if defined(APP_USBD_ENABLED) && APP_USBD_ENABLED

  while (app_usbd_event_queue_process())
  {
    /* Nothing to do */
  }
#endif
}

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const *p_ble_evt, void *p_context)
{
  ret_code_t err_code;
  NRF_LOG_INFO("Received BLE event: %d", p_ble_evt->header.evt_id);
  switch (p_ble_evt->header.evt_id)
  {
  case BLE_GAP_EVT_CONNECTED: // 16
    NRF_LOG_INFO("Connected.");
    // APP_ERROR_CHECK(err_code);
    //  err_code = nrf_ble_qwr_conn_handle_assign(&m_qwr, m_conn_handle);
    // APP_ERROR_CHECK(err_code);
    //  err_code = pm_conn_secure(p_ble_evt->evt.gap_evt.conn_handle, false);
    //  if (err_code != NRF_ERROR_BUSY)
    //  {
    //    APP_ERROR_CHECK(err_code);
    //  }
    NRF_LOG_INFO("Connection handle: 0x%x", p_ble_evt->evt.gap_evt.conn_handle);
    on_connected(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GAP_EVT_DISCONNECTED: // 17
    NRF_LOG_INFO("Disconnected, reason %d.",
                 p_ble_evt->evt.gap_evt.params.disconnected.reason);
    break;

  case BLE_GAP_EVT_PHY_UPDATE_REQUEST: // 33
  {
    NRF_LOG_INFO("PHY update request.");
    // ble_gap_phys_t const phys =
    //     {
    //         .rx_phys = BLE_GAP_PHY_AUTO,
    //         .tx_phys = BLE_GAP_PHY_AUTO,
    //     };
    // err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
    // APP_ERROR_CHECK(err_code);
    on_phy_requeset(&(p_ble_evt->evt.gap_evt));
  }
  break;

  case BLE_GAP_EVT_TIMEOUT: // 27
    on_timeout(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GATTC_EVT_TIMEOUT: // 59
    // Disconnect on GATT Client timeout event.
    NRF_LOG_DEBUG("GATT Client Timeout.");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                     BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTS_EVT_TIMEOUT: // 86
    // Disconnect on GATT Server timeout event.
    NRF_LOG_DEBUG("GATT Server Timeout.");
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                     BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GAP_EVT_SEC_PARAMS_REQUEST: // 19
    NRF_LOG_DEBUG("BLE_GAP_EVT_SEC_PARAMS_REQUEST");
    break;

  case BLE_GAP_EVT_AUTH_KEY_REQUEST: // 23
    NRF_LOG_INFO("BLE_GAP_EVT_AUTH_KEY_REQUEST");
    break;

  case BLE_GAP_EVT_LESC_DHKEY_REQUEST: // 24
    NRF_LOG_INFO("BLE_GAP_EVT_LESC_DHKEY_REQUEST");
    break;

  case BLE_GAP_EVT_AUTH_STATUS: // 25
    NRF_LOG_INFO("BLE_GAP_EVT_AUTH_STATUS: status=0x%x bond=0x%x lv4: %d kdist_own:0x%x kdist_peer:0x%x",
                 p_ble_evt->evt.gap_evt.params.auth_status.auth_status,
                 p_ble_evt->evt.gap_evt.params.auth_status.bonded,
                 p_ble_evt->evt.gap_evt.params.auth_status.sm1_levels.lv4,
                 *((uint8_t *)&p_ble_evt->evt.gap_evt.params.auth_status.kdist_own),
                 *((uint8_t *)&p_ble_evt->evt.gap_evt.params.auth_status.kdist_peer));
    break;

  case BLE_GAP_EVT_ADV_REPORT: // 29
  {
    // char addr_str[7];
    // memset(addr_str, 0, sizeof(addr_str));
    // ble_address_to_string_convert(p_ble_evt->evt.gap_evt.params.adv_report.peer_addr, addr_str);
    // addr_str[6] = '\0';

    // NRF_LOG_INFO("BLE_GAP_EVT_ADV_REPORT_Connectable: %d, directed: %d", p_ble_evt->evt.gap_evt.params.adv_report.type.connectable, p_ble_evt->evt.gap_evt.params.adv_report.type.directed);
    // NRF_LOG_INFO("BLE_GAP_EVT_ADV_REPORT: addr=%s, rssi=%d", addr_str,
    //              p_ble_evt->evt.gap_evt.params.adv_report.rssi);
    // uint32_t err_code;
    // memset(mp_data, 0, sizeof(mp_data));
    // m_adv_report_data.len = sizeof(mp_data);
    // m_adv_report_data.p_data = mp_data;
    // err_code = sd_ble_gap_scan_start(NULL, &m_adv_report_data);
    // if (err_code != NRF_SUCCESS)
    //   NRF_LOG_DEBUG("sd_ble_gap_scan_start: 0x%4x", err_code);
    on_adv_report(&(p_ble_evt->evt.gap_evt));
    break;
  }

  case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP: // 48
    on_service_discovery_response(&p_ble_evt->evt.gattc_evt);
    break;

  case BLE_GATTC_EVT_CHAR_DISC_RSP: // 50
    on_characteristic_discovery_response(&(p_ble_evt->evt.gattc_evt));
    break;
  case BLE_GATTC_EVT_DESC_DISC_RSP: // 51
    on_descriptor_discovery_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_WRITE_RSP: // 56
    on_write_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_HVX: // 57
    on_hvx(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST: // 31
    on_conn_params_update_request(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GAP_EVT_CONN_PARAM_UPDATE: // 18
    const ble_gap_conn_params_t *conn_params;

    conn_params = &p_ble_evt->evt.gap_evt.params.conn_param_update.conn_params;

    // printf("Connection parameters updated. New parameters:\n");
    // printf("Connection interval : %d [ms]\n",
    //        (int)(conn_params->min_conn_interval * 1.25));
    // printf("Slave latency : %d\n", conn_params->slave_latency);
    // printf("Supervision timeout : %d [s]\n",
    //        (int)(conn_params->conn_sup_timeout / 100));
    ble_usb_add_string("Conn Params = Min Conn Interval = ", 34);
    ble_usb_add_value_uint16(conn_params->min_conn_interval, 0);
    ble_usb_add_string(", Max Conn Interval = ", 22);
    ble_usb_add_value_uint16(conn_params->max_conn_interval, 0);
    ble_usb_add_string(", Slave Latency = ", 18);
    ble_usb_add_value_uint16(conn_params->slave_latency, 0);
    ble_usb_add_string(", Conn Superversion Timeout = ", 30);
    ble_usb_add_value_uint16(conn_params->conn_sup_timeout, 0);
    ble_usb_add_string("\n", 1);
    ble_usb_printf();
    break;

  case BLE_GATTC_EVT_READ_RSP: // 54
    on_read_response(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GATTC_EVT_CHAR_VALS_READ_RSP: // 55
    on_char_val_by_uuid_read_rsp(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GATTC_EVT_CHAR_VAL_BY_UUID_READ_RSP: // 53
    on_char_val_by_uuid_read_rsp(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GATTC_EVT_ATTR_INFO_DISC_RSP: // 52
    on_attr_info_disc(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST: // 85
    on_exchange_mtu_request(&(p_ble_evt->evt.gatts_evt));
    break;

  case BLE_GATTC_EVT_EXCHANGE_MTU_RSP: // 58
    on_exchange_mtu_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GAP_EVT_DATA_LENGTH_UPDATE: // 36
    // printf("Maximum radio packet length updated to %d bytes.\n",
    //        p_ble_evt->evt.gap_evt.params.data_length_update.effective_params
    //            .max_tx_octets);
    BLE_USB_PRINT_MSG_NUM("Maximum Radio Packet Length Updated To ", p_ble_evt->evt.gap_evt.params.data_length_update.effective_params.max_tx_octets, 0);
    break;

  case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST: // 35
    sd_ble_gap_data_length_update(m_connection_handle, NULL, NULL);
    break;

  case BLE_GATTS_EVT_SYS_ATTR_MISSING: // 82
    sd_ble_gatts_sys_attr_set(m_connection_handle, NULL, 0, 0);
    break;

  case BLE_GATTS_EVT_WRITE: // 80
    on_gatts_write(p_ble_evt->evt.gatts_evt.params.write);
    break;

  default:
    // No implementation needed.
    break;
  }
  uint16_t len = 0;
  sd_ble_evt_get(m_ble_evt_buf, &len);
}

static void ble_stack_init(void)
{
  ret_code_t err_code;

  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);
  NRF_LOG_DEBUG("nrf_sdh_enable_request: 0x%4x", err_code);

  // Configure the BLE stack using the default settings.
  // Fetch the start address of the application RAM.
  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(m_config_id, &ram_start);
  NRF_LOG_DEBUG("RAM start: 0x%08x", ram_start);
  NRF_LOG_DEBUG("nrf_sdh_ble_default_cfg_set: 0x%4x", err_code);
  APP_ERROR_CHECK(err_code);

  // Enable BLE stack.
  err_code = nrf_sdh_ble_enable(&ram_start);
  NRF_LOG_DEBUG("nrf_sdh_ble_enable: 0x%4x", err_code);
  NRF_LOG_DEBUG("Ram start: 0x%08x", ram_start);
  APP_ERROR_CHECK(err_code);

  // Register a handler for BLE events.
  NRF_SDH_BLE_OBSERVER(m_ble_observer, 3, ble_evt_handler, NULL);
  NRF_LOG_DEBUG("m_ble_observer: 0x%4x", &m_ble_observer);
  if (m_ble_observer.handler != ble_evt_handler)
  {
    NRF_LOG_DEBUG("m_ble_observer.handler: 0x%4x", m_ble_observer.handler);
    NRF_LOG_DEBUG("ble_evt_handler: 0x%4x", ble_evt_handler);
  }
}

/**@brief   Sleep until an event is received. */
static void power_manage(void)
{
#ifdef SOFTDEVICE_PRESENT
  (void)sd_app_evt_wait();
#else
  __WFE();
#endif
}

/* Flag to check fds initialization. */
static bool volatile m_fds_initialized;

/**@brief   Wait for fds to initialize. */
static void wait_for_fds_ready(void)
{
  while (!m_fds_initialized)
  {
    power_manage();
  }
}

#define SN_SIZE 12
char SN[SN_SIZE] = {0};

static fds_record_t const m_dummy_record =
    {
        .file_id = CONFIG_FILE,
        .key = CONFIG_REC_KEY,
        .data.p_data = SN,
        .data.length_words = 12,
};

/* Keep track of the progress of a delete_all operation. */
static struct
{
  bool delete_next; //!< Delete next record.
  bool pending;     //!< Waiting for an fds FDS_EVT_DEL_RECORD event, to delete the next record.
} m_delete_all;

// fds_evt_handler
static void fds_evt_handler(fds_evt_t const *p_evt)
{
  if (p_evt->result == NRF_SUCCESS)
  {
    //        NRF_LOG_GREEN("Event: %s received (NRF_SUCCESS)",
    //                      fds_evt_str[p_evt->id]);
  }
  else
  {
    //        NRF_LOG_GREEN("Event: %s received (%s)",
    //                      fds_evt_str[p_evt->id],
    //                      fds_err_str(p_evt->result));
  }

  switch (p_evt->id)
  {
  case FDS_EVT_INIT:
    if (p_evt->result == NRF_SUCCESS)
    {
      m_fds_initialized = true;
    }
    break;

  case FDS_EVT_WRITE:
  {
    if (p_evt->result == NRF_SUCCESS)
    {
      NRF_LOG_INFO("Record ID:\t0x%04x", p_evt->write.record_id);
      NRF_LOG_INFO("File ID:\t0x%04x", p_evt->write.file_id);
      NRF_LOG_INFO("Record key:\t0x%04x", p_evt->write.record_key);
    }
  }
  break;

  case FDS_EVT_DEL_RECORD:
  {
    if (p_evt->result == NRF_SUCCESS)
    {
      NRF_LOG_INFO("Record ID:\t0x%04x", p_evt->del.record_id);
      NRF_LOG_INFO("File ID:\t0x%04x", p_evt->del.file_id);
      NRF_LOG_INFO("Record key:\t0x%04x", p_evt->del.record_key);
    }
    m_delete_all.pending = false;
  }
  break;

  default:
    break;
  }
}

/**
 * @brief Enable power USB detection
 *
 * Configure if example supports USB port connection
 */
#ifndef USBD_POWER_DETECTION
#define USBD_POWER_DETECTION true
#endif

static void usbd_user_ev_handler(app_usbd_event_type_t event)
{
  switch (event)
  {
  case APP_USBD_EVT_DRV_SUSPEND:
    //            bsp_board_led_off(LED_USB_RESUME);
    break;
  case APP_USBD_EVT_DRV_RESUME:
    //            bsp_board_led_on(LED_USB_RESUME);
    break;
  case APP_USBD_EVT_STARTED:
    break;
  case APP_USBD_EVT_STOPPED:
    app_usbd_disable();
    bsp_board_leds_off();
    break;
  case APP_USBD_EVT_POWER_DETECTED:
    NRF_LOG_INFO("USB power detected");

    if (!nrf_drv_usbd_is_enabled())
    {
      app_usbd_enable();
    }
    break;
  case APP_USBD_EVT_POWER_REMOVED:
    NRF_LOG_INFO("USB power removed");
    app_usbd_stop();
    break;
  case APP_USBD_EVT_POWER_READY:
    NRF_LOG_INFO("USB ready");
    app_usbd_start();
    break;
  default:
    break;
  }
}

void usb_init()
{
  ret_code_t ret;
  static const app_usbd_config_t usbd_config = {
      .ev_state_proc = usbd_user_ev_handler};

  ret = nrf_drv_clock_init();
  APP_ERROR_CHECK(ret);

  nrf_drv_clock_lfclk_request(NULL);

  while (!nrf_drv_clock_lfclk_is_running())
  {
  }

  // ret = app_timer_init();
  // APP_ERROR_CHECK(ret);

  app_usbd_serial_num_generate();

  ret = app_usbd_init(&usbd_config);
  APP_ERROR_CHECK(ret);

  // Enable DFU Trigger library over USB (should happen right after usbd_init)
  //    nrf_dfu_trigger_usb_init();

  // USBD CDC ACM example started
  app_usbd_class_inst_t const *class_cdc_acm = app_usbd_cdc_acm_class_inst_get(usb_uart);
  ret = app_usbd_class_append(class_cdc_acm);
  APP_ERROR_CHECK(ret);

  if (USBD_POWER_DETECTION)
  {
    ret = app_usbd_power_events_enable();
    APP_ERROR_CHECK(ret);
  }
  else
  {
    // No USB power detection enabled. Starting USB now
    app_usbd_enable();
    app_usbd_start();
  }
}

void ble_event_get_with_exe(void)
{
  uint16_t event_len = BLE_EVT_BUF_SIZE;
  uint32_t err_code;

  //  SoftDevice ȡ¼
  while ((err_code = sd_ble_evt_get(m_ble_evt_buf, &event_len)) == NRF_SUCCESS)
  {
    ble_evt_t *p_ble_evt = (ble_evt_t *)m_ble_evt_buf;
    ble_evt_handler(p_ble_evt, NULL);
    NRF_LOG_INFO("Received BLE event with length %d", event_len);
  }
}

/**@brief Main function of the connectivity application. */
int main(void)
{
  uint32_t err_code = NRF_SUCCESS;

  APP_ERROR_CHECK(NRF_LOG_INIT(NULL));

  NRF_LOG_DEFAULT_BACKENDS_INIT();

  NRF_LOG_INFO("BLE connectivity started");

  // bsp_board_init(BSP_INIT_LEDS);

#if (defined(SER_PHY_HCI_DEBUG_ENABLE) || defined(SER_PHY_DEBUG_APP_ENABLE))
  debug_init(NULL);
#endif

  /* Force constant latency mode to control SPI slave timing */
  NRF_POWER->TASKS_CONSTLAT = 1;

  /* Initialize scheduler queue. */
  // lint -save -e666 -e587
  APP_SCHED_INIT(SER_CONN_SCHED_MAX_EVENT_DATA_SIZE, SER_CONN_SCHED_QUEUE_SIZE);
  // lint -restore

  /* Initialize SoftDevice.
   * SoftDevice Event IRQ is not scheduled but immediately copies BLE events to the application
   * scheduler queue */

  // err_code = nrf_drv_clock_init();
  // if ((err_code != NRF_SUCCESS) &&
  //     (err_code != NRF_ERROR_MODULE_ALREADY_INITIALIZED))
  // {
  //     APP_ERROR_CHECK(err_code);
  // }

  // nrf_drv_clock_hfclk_request(NULL);
  // while (!nrf_drv_clock_hfclk_is_running())
  // {}

#if defined(APP_USBD_ENABLED) && APP_USBD_ENABLED
  // usbd_init();
#endif

  /* Open serialization HAL Transport layer and subscribe for HAL Transport events. */
  // err_code = ser_hal_transport_open(ser_conn_hal_transport_event_handle);
  // APP_ERROR_CHECK(err_code);

#if defined(APP_USBD_ENABLED) && APP_USBD_ENABLED
  // usbd_enable();
#endif

  // err_code = nrf_sdh_enable_request();
  // APP_ERROR_CHECK(err_code);

  err_code = app_timer_init();
  APP_ERROR_CHECK(err_code);

  /* Register first to receive an event when initialization is complete. */
  (void)fds_register(fds_evt_handler);

  NRF_LOG_INFO("Initializing fds...");

  err_code = fds_init();
  APP_ERROR_CHECK(err_code);

  /* Wait for fds to initialize. */
  wait_for_fds_ready();

  usb_init();

  ble_stack_init();
  //  m_scan_start();

  // ser_conn_on_no_mem_handler_set(on_idle);

  /* Enter main loop. */
  for (;;)
  {
    // /* Process SoftDevice events. */
    // app_sched_execute();

    // CRITICAL_REGION_ENTER();
    // if (nrf_sdh_is_suspended())
    // {
    //     // Resume pulling new events if queue utilization drops below 50%.
    //     if (app_sched_queue_space_get() > (SER_CONN_SCHED_QUEUE_SIZE >> 1))
    //     {
    //         nrf_sdh_resume();
    //     }
    // }
    // CRITICAL_REGION_EXIT();

    // /* Process received packets.
    //  * We can NOT add received packets as events to the application scheduler queue because
    //  * received packets have to be processed before SoftDevice events but the scheduler queue
    //  * does not have priorities. */
    // err_code = ser_conn_rx_process();
    // APP_ERROR_CHECK(err_code);

    on_idle();
    ble_event_get_with_exe();
  }
}
/** @} */

#define READ_SIZE 1

static char m_rx_buffer[READ_SIZE];
char Rx_data[1024] = {0};

/**
 * @brief User event handler @ref app_usbd_cdc_acm_user_ev_handler_t (headphones)
 * */
void cdc_acm_user_ev_handler(app_usbd_class_inst_t const *p_inst,
                             app_usbd_cdc_acm_user_event_t event)
{
  app_usbd_cdc_acm_t const *p_cdc_acm = app_usbd_cdc_acm_class_get(p_inst);
  switch (event)
  {
  case APP_USBD_CDC_ACM_USER_EVT_PORT_OPEN:
  {
    //							bsp_board_led_on(LED_CDC_ACM_OPEN);

    /*Setup first transfer*/
    ret_code_t ret = app_usbd_cdc_acm_read(usb_uart,
                                           m_rx_buffer,
                                           READ_SIZE);
    UNUSED_VARIABLE(ret);
    break;
  }
  case APP_USBD_CDC_ACM_USER_EVT_PORT_CLOSE:
    //							bsp_board_led_off(LED_CDC_ACM_OPEN);
    break;
  case APP_USBD_CDC_ACM_USER_EVT_TX_DONE:
    //							bsp_board_led_invert(LED_CDC_ACM_TX);
    break;
  case APP_USBD_CDC_ACM_USER_EVT_RX_DONE:
  {
    ret_code_t ret;
    int index = 0; // յݵܳ
    //							char test_string[512]={0,0,0,0,0,0,48};
    memset(Rx_data, 0, 1024); // ݻ
    // NRF_LOG_INFO("Bytes waiting: %d", app_usbd_cdc_acm_bytes_stored(p_cdc_acm));
    ret = NRF_SUCCESS;
    do
    {
      Rx_data[index] = m_rx_buffer[0];
      index++;
      /* Fetch data until internal buffer is empty */
      ret = app_usbd_cdc_acm_read(usb_uart,
                                  m_rx_buffer,
                                  READ_SIZE);
    } while (ret == NRF_SUCCESS);
    if (Rx_data == strstr(Rx_data, ">SN_WR:")) // յ>SN_WR:ͷ
    {
      //			nrf_sdh_disable_request();
      sscanf(Rx_data, ">SN_WR:%s", SN);
      //			usb_printf(SN);
      NRF_LOG_INFO("uart sacn SN:%s", SN);
      fds_record_desc_t desc = {0};
      fds_find_token_t tok = {0};
      fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok); // ѯṩFile ID and KeyǷʷд룬ûзʧܣ򷵻سɹdesctokждĵһ
      // flash write
      fds_record_delete(&desc);                 // ɾһ¼
      fds_record_write(&desc, &m_dummy_record); // յSNдflash
      nrf_sdh_enable_request();
    }
    else if (Rx_data == strstr(Rx_data, ">SN_RD")) // յ>SN_RDڷflashд洢SN
    {
      //			nrf_sdh_disable_request();
      fds_gc(); // Function for running garbage collection,ձfds_record_deleteɾ¼ڴռ䣬֮fds_record_open
      fds_record_desc_t desc = {0};
      fds_find_token_t tok = {0};
      fds_record_find(CONFIG_FILE, CONFIG_REC_KEY, &desc, &tok); // ѯṩFile ID and KeyǷʷд룬ûзʧܣ򷵻سɹdesctokждĵһ
      // flash read
      /* A config file is in flash. Let's update it. */
      fds_flash_record_t config = {0};

      /* Open the record and read its contents. */
      ret_code_t rc = fds_record_open(&desc, &config);
      // usb_printf((char *)config.p_data);

      memcpy(SN, config.p_data, SN_SIZE);

      fds_record_close(&desc);

      fds_record_update(&desc, &m_dummy_record); // ¼¼

      usb_printf_len(SN, 12); // ڷflashд洢SN

      nrf_sdh_enable_request();
    }
    else if (Rx_data == strstr(Rx_data, ">SD:")) // յ>SD:ͷ
    {
      get_cmd(Rx_data, index);
    }
    break;
  }
  default:
    break;
  }
}
